@@ -37,6 +37,7 @@
 #include "ftw.h"
 #include "hash.h"
 #include "human.h"
+#include "mmap-stack.h"
 #include "quote.h"
 #include "quotearg.h"
 #include "same.h"
@@ -73,16 +74,20 @@ static Hash_table *htab;
 /* Name under which this program was invoked.  */
 char *program_name;
 
-/* If nonzero, display counts for all files, not just directories. */
+/* If nonzero, display counts for all files, not just directories.  */
 static int opt_all = 0;
 
-/* If nonzero, count each hard link of files with multiple links. */
+/* If nonzero, rather than using the disk usage of each file,
+   use the apparent size (a la stat.st_size).  */
+static int apparent_size = 0;
+
+/* If nonzero, count each hard link of files with multiple links.  */
 static int opt_count_all = 0;
 
-/* If nonzero, print a grand total at the end. */
+/* If nonzero, print a grand total at the end.  */
 static int print_totals = 0;
 
-/* If nonzero, do not add sizes of subdirectories. */
+/* If nonzero, do not add sizes of subdirectories.  */
 static int opt_separate_dirs = 0;
 
 /* If nonzero, dereference symlinks that are command line arguments.
@@ -106,7 +111,7 @@ static uintmax_t output_block_size;
 /* File name patterns to exclude.  */
 static struct exclude *exclude;
 
-/* Grand total size of all args, in units of ST_NBLOCKSIZE-byte blocks. */
+/* Grand total size of all args, in bytes. */
 static uintmax_t tot_size = 0;
 
 /* In some cases, we have to append `/.' or just `.' to an argument
@@ -132,13 +137,15 @@ int G_fail;
    non-character as a pseudo short option, starting with CHAR_MAX + 1.  */
 enum
 {
-  EXCLUDE_OPTION = CHAR_MAX + 1,
+  APPARENT_SIZE_OPTION = CHAR_MAX + 1,
+  EXCLUDE_OPTION,
   MAX_DEPTH_OPTION
 };
 
 static struct option const long_options[] =
 {
   {"all", no_argument, NULL, 'a'},
+  {"apparent-size", no_argument, NULL, APPARENT_SIZE_OPTION},
   {"block-size", required_argument, 0, 'B'},
   {"bytes", no_argument, NULL, 'b'},
   {"count-links", no_argument, NULL, 'l'},
@@ -178,8 +185,12 @@ Mandatory arguments to long options are mandatory for short options too.\n\
 "), stdout);
       fputs (_("\
   -a, --all             write counts for all files, not just directories\n\
+      --apparent-size   print apparent sizes, rather than disk usage; although\n\
+                          the apparent size is usually smaller, it may be\n\
+                          larger due to holes in (`sparse') files, internal\n\
+                          fragmentation, indirect blocks, and the like\n\
   -B, --block-size=SIZE use SIZE-byte blocks\n\
-  -b, --bytes           print size in bytes\n\
+  -b, --bytes           equivalent to `--apparent-size --block-size=1'\n\
   -c, --total           produce a grand total\n\
   -D, --dereference-args  dereference FILEs that are symbolic links\n\
 "), stdout);
@@ -277,30 +288,24 @@ hash_init (void)
     xalloc_die ();
 }
 
-/* Print N_BLOCKS.  NBLOCKS is the number of
-   ST_NBLOCKSIZE-byte blocks; convert it to a readable value before
-   printing.  */
+/* Print N_BYTES.  Convert it to a readable value before printing.  */
 
 static void
-print_only_size (uintmax_t n_blocks)
+print_only_size (uintmax_t n_bytes)
 {
   char buf[LONGEST_HUMAN_READABLE + 1];
-  fputs (human_readable (n_blocks, buf, human_output_opts,
-			 ST_NBLOCKSIZE, output_block_size), stdout);
+  fputs (human_readable (n_bytes, buf, human_output_opts,
+			 1, output_block_size), stdout);
 }
 
-/* Print N_BLOCKS followed by STRING on a line.  NBLOCKS is the number of
-   ST_NBLOCKSIZE-byte blocks; convert it to a readable value before
-   printing.  */
+/* Print N_BYTES followed by STRING on a line.
+   Convert N_BYTES to a readable value before printing.  */
 
 static void
-print_size (uintmax_t n_blocks, const char *string)
+print_size (uintmax_t n_bytes, const char *string)
 {
-  char buf[LONGEST_HUMAN_READABLE + 1];
-  printf ("%s\t%s\n",
-	  human_readable (n_blocks, buf, human_output_opts,
-			  ST_NBLOCKSIZE, output_block_size),
-	  string);
+  print_only_size (n_bytes);
+  printf ("\t%s\n", string);
   fflush (stdout);
 }
 
@@ -314,19 +319,15 @@ process_file (const char *file, const struct stat *sb, int file_type,
 	      struct FTW *info)
 {
   size_t size;
-  size_t s;
+  size_t size_including_subdirs;
+  size_t size_to_propagate_to_parent;
   static int first_call = 1;
   static size_t prev_level;
   static size_t n_alloc;
-  /* The sum of the st_size values of all entries in the single directory
-     at the corresponding level.  Although this does include the st_size
-     corresponding to each subdirectory, it does not include the size of
-     any file in a subdirectory.  */
-  static uintmax_t *sum_ent;
+  static uintmax_t *sum;
 
-  /* The sum of the sizes of all entries in the hierarchy at or below the
-     directory at the specified level.  */
-  static uintmax_t *sum_subdir;
+  /* Always define info->skip before returning.  */
+  info->skip = excluded_filename (exclude, file + info->base);
 
   switch (file_type)
     {
@@ -359,9 +360,14 @@ process_file (const char *file, const struct stat *sb, int file_type,
       break;
     }
 
+  /* If this is the first (pre-order) encounter with a directory,
+     return right away.  */
+  if (file_type == FTW_DPRE)
+    return 0;
+
   /* If the file is being excluded or if it has already been counted
      via a hard link, then don't let it contribute to the sums.  */
-  if ((info->skip = excluded_filename (exclude, file + info->base))
+  if (info->skip
       || (!opt_count_all
 	  && 1 < sb->st_nlink
 	  && hash_ins (sb->st_ino, sb->st_dev)))
@@ -369,23 +375,19 @@ process_file (const char *file, const struct stat *sb, int file_type,
       /* Note that we must not simply return here.
 	 We still have to update prev_level and maybe propagate
 	 some sums up the hierarchy.  */
-      s = size = 0;
+      size = 0;
     }
   else
     {
-      s = size = ST_NBLOCKS (*sb);
+      size = (apparent_size
+	      ? sb->st_size
+	      : ST_NBLOCKS (*sb) * ST_NBLOCKSIZE);
     }
 
-  /* If this is the first (pre-order) encounter with a directory,
-     return right away.  */
-  if (file_type == FTW_DPRE)
-    return 0;
-
   if (first_call)
     {
       n_alloc = info->level + 10;
-      sum_ent = XCALLOC (uintmax_t, n_alloc);
-      sum_subdir = XCALLOC (uintmax_t, n_alloc);
+      sum = XCALLOC (uintmax_t, n_alloc);
     }
   else
     {
@@ -397,18 +399,29 @@ process_file (const char *file, const struct stat *sb, int file_type,
       if (n_alloc <= (size_t) info->level)
 	{
 	  n_alloc = info->level * 2;
-	  sum_ent = XREALLOC (sum_ent, uintmax_t, n_alloc);
-	  sum_subdir = XREALLOC (sum_subdir, uintmax_t, n_alloc);
+	  sum = XREALLOC (sum, uintmax_t, n_alloc);
 	}
     }
 
+  size_to_propagate_to_parent = size_including_subdirs = size;
+
   if (! first_call)
     {
       if ((size_t) info->level == prev_level)
 	{
 	  /* This is usually the most common case.  Do nothing.  */
 	}
-      else if ((size_t) info->level < prev_level)
+      else if ((size_t) info->level > prev_level)
+	{
+	  /* Descending the hierarchy.
+	     Clear the accumulators for *all* levels between prev_level
+	     and the current one.  The depth may change dramatically,
+	     e.g., from 1 to 10.  */
+	  int i;
+	  for (i = prev_level + 1; i <= info->level; i++)
+	    sum[i] = 0;
+	}
+      else /* info->level < prev_level */
 	{
 	  /* Ascending the hierarchy.
 	     nftw processes a directory only after all entries in that
@@ -417,31 +430,18 @@ process_file (const char *file, const struct stat *sb, int file_type,
 	     Here, the current level is always one smaller than the
 	     previous one.  */
 	  assert ((size_t) info->level == prev_level - 1);
-	  s += sum_ent[prev_level];
-	  if (!opt_separate_dirs)
-	    s += sum_subdir[prev_level];
-	  sum_subdir[info->level] += (sum_ent[prev_level]
-				      + sum_subdir[prev_level]);
-	}
-      else /* info->level > prev_level */
-	{
-	  /* Descending the hierarchy.
-	     Clear the accumulators for *all* levels between prev_level
-	     and the current one.  The depth may change dramatically,
-	     e.g., from 1 to 10.  */
-	  int i;
-	  for (i = prev_level + 1; i <= info->level; i++)
-	    {
-	      sum_ent[i] = 0;
-	      sum_subdir[i] = 0;
-	    }
+	  size_to_propagate_to_parent = size_including_subdirs
+	    = size + sum[prev_level];
 	}
     }
 
+  if (opt_separate_dirs)
+    size_to_propagate_to_parent = 0;
+
   prev_level = info->level;
   first_call = 0;
 
-  sum_ent[info->level] += size;
+  sum[info->level] += size_to_propagate_to_parent;
 
   /* Even if this directory was unreadable or we couldn't chdir into it,
      do let its size contribute to the total, ... */
@@ -457,7 +457,7 @@ process_file (const char *file, const struct stat *sb, int file_type,
 		     (info->level <= max_depth || info->level == 0))
       || ((opt_all && info->level <= max_depth) || info->level == 0))
     {
-      print_only_size (s);
+      print_only_size (size_including_subdirs);
       fputc ('\t', stdout);
       if (arg_length)
 	{
@@ -497,7 +497,8 @@ is_symlink_to_dir (char const *file)
    FTW_FLAGS controls how nftw works.
    Return nonzero upon error.  */
 
-static int
+static void du_files (char **files, int ftw_flags) ATTRIBUTE_NORETURN;
+static void
 du_files (char **files, int ftw_flags)
 {
   int fail = 0;
@@ -544,7 +545,7 @@ du_files (char **files, int ftw_flags)
   if (print_totals)
     print_size (tot_size, _("total"));
 
-  return fail;
+  exit (fail || G_fail ? EXIT_FAILURE : EXIT_SUCCESS);
 }
 
 int
@@ -591,7 +592,12 @@ main (int argc, char **argv)
 	  opt_all = 1;
 	  break;
 
+	case APPARENT_SIZE_OPTION:
+	  apparent_size = 1;
+	  break;
+
 	case 'b':
+	  apparent_size = 1;
 	  human_output_opts = 0;
 	  output_block_size = 1;
 	  break;
@@ -716,6 +722,5 @@ main (int argc, char **argv)
   /* Initialize the hash structure for inode numbers.  */
   hash_init ();
 
-  exit (du_files (files, ftw_flags) || G_fail
-	? EXIT_FAILURE : EXIT_SUCCESS);
+  RUN_WITH_BIG_STACK_2 (du_files, files, ftw_flags);
 }
