@@ -148,23 +148,23 @@ struct Word
 
 /* Forward declarations.  */
 
-static void set_prefix PARAMS ((char *p));
-static void fmt PARAMS ((FILE *f));
-static bool get_paragraph PARAMS ((FILE *f));
-static int get_line PARAMS ((FILE *f, int c));
-static int get_prefix PARAMS ((FILE *f));
-static int get_space PARAMS ((FILE *f, int c));
-static int copy_rest PARAMS ((FILE *f, int c));
-static bool same_para PARAMS ((int c));
-static void flush_paragraph PARAMS ((void));
-static void fmt_paragraph PARAMS ((void));
-static void check_punctuation PARAMS ((WORD *w));
-static COST base_cost PARAMS ((WORD *this));
-static COST line_cost PARAMS ((WORD *next, int len));
-static void put_paragraph PARAMS ((WORD *finish));
-static void put_line PARAMS ((WORD *w, int indent));
-static void put_word PARAMS ((WORD *w));
-static void put_space PARAMS ((int space));
+static void set_prefix (char *p);
+static void fmt (FILE *f);
+static bool get_paragraph (FILE *f);
+static int get_line (FILE *f, int c);
+static int get_prefix (FILE *f);
+static int get_space (FILE *f, int c);
+static int copy_rest (FILE *f, int c);
+static bool same_para (int c);
+static void flush_paragraph (void);
+static void fmt_paragraph (void);
+static void check_punctuation (WORD *w);
+static COST base_cost (WORD *this);
+static COST line_cost (WORD *next, int len);
+static void put_paragraph (WORD *finish);
+static void put_line (WORD *w, int indent);
+static void put_word (WORD *w);
+static void put_space (int space);
 
 /* The name this program was run with.  */
 const char *program_name;
