@@ -46,10 +46,10 @@ gid_t getegid ();
 
 int getugroups ();
 
-static void print_user PARAMS ((uid_t uid));
-static void print_group PARAMS ((gid_t gid));
-static void print_group_list PARAMS ((const char *username));
-static void print_full_info PARAMS ((const char *username));
+static void print_user (uid_t uid);
+static void print_group (gid_t gid);
+static void print_group_list (const char *username);
+static void print_full_info (const char *username);
 
 /* The name this program was run with. */
 char *program_name;
@@ -308,7 +308,8 @@ print_group_list (const char *username)
     GETGROUPS_T *groups;
     register int i;
 
-    if (xgetgroups (username, pwd ? pwd->pw_gid : -1, &n_groups, &groups))
+    if (xgetgroups (username, (pwd ? pwd->pw_gid : (gid_t) -1),
+		    &n_groups, &groups))
       {
 	++problems;
 	return;
@@ -373,7 +374,8 @@ print_full_info (const char *username)
     GETGROUPS_T *groups;
     register int i;
 
-    if (xgetgroups (username, pwd ? pwd->pw_gid : -1, &n_groups, &groups))
+    if (xgetgroups (username, (pwd ? pwd->pw_gid : (gid_t) -1),
+		    &n_groups, &groups))
       {
 	++problems;
 	return;
